_SCHEMA_DEFAULT = {'title': None, 'multipleOf': None, 'maximum': None, 'exclusiveMaximum': None, 'minimum': None, 'exclusiveMinimum': None, 'maxLength': None, 'minLength': None, 'pattern': None, 'maxItems': None, 'minItems': None, 'uniqueItems': None, 'maxProperties': None, 'minProperties': None, 'required': None, 'enum': None, 'allOf': None, 'oneOf': None, 'anyOf': None, 'items': None, 'properties': None, 'additionalProperties': None, 'description': None, 'default': None, 'nullable': None, 'discriminator': None, 'readOnly': None, 'writeOnly': None, 'example': None, 'externalDocs': None, 'deprecated': None, 'xml': None, '_format': None, '_type': None, '_not': None, '_x': None}